# -*- python -*-
# $RCSfile: worker.py,v $
# $Revision: 1.72 $
# $Author: reida $
# $Date: 2005/08/04 18:31:19 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


## A worker is a wrapped generalized callback (from a menuitem) that will
## run in a separate thread. This callback is monitored through progress
## bars, and gtk progress bars.
##

from oof2.SWIG.common import lock
from oof2.SWIG.common.IO import progressbar
from oof2.SWIG.common.IO import threadstate
from oof2.common import debug
from oof2.common import mainthread
from oof2.common import parallel_enable
from oof2.common import thread_enable
from oof2.common import utils
from oof2.common.IO import reporter
from oof2.common.IO import threadmanager

import sys
import threading
import time

if parallel_enable.query()==1:
    from oof2.common.IO.socket2me import pipeToSocket

# Worker classes.  The basic highest-level interface is that workers
# must have a "start" routine, which is called from the menu item
# after the worker is created, and it must also have a
# "getProgressBar" routine which takes a progress bar type.  The
# second one is required because the type of progress bar required is
# not generally known until run-time within the callback.

# Threading is handled with an extra layer of indirection -- the
# "ThreadedWorker" and subclasses are actually wrappers, which contain
# a "ThreadedWorkerCore" object.  These wrapper objects call their
# core object's "start" routine, and then manage the main thread,
# which may want to do complicated things before (or instead of)
# returning control to the caller.  The TextThreadedWorker draws the
# progress bar for the command.  The GUIThreadedWorkerBlock object,
# used in GUI mode, blocks until the just-launched thread terminates.
# This is useful when running scripts, for instance, when each script
# menu item (threadable or otherwise) runs on a subthread, but must
# not run concurrently with previous script commands.  This technique
# allows the GUI to be responsive and a script-reading progress bar to
# be presented during the read.

# Parallelization is fairly straightforward at this level -- menu
# items which are parallelized require a worker which is a subclass of
# ParallelWork, which itself only ensures that all the back-end
# machines receive the same menu item.

# Flag to indicate how to handle exceptions.  If set to 0, workers
# should catch exceptions and send them to the reporter.  If nonzero,
# workers should allow excpetions to propagate to the caller.  In the
# case of threaded workers, the core should propagate the exception
# data to the calling worker, which should re-raise it.
propagate_exceptions=0

class NonThreadedWorker:
    def __init__(self, menuitem, kwargs):
        self.menuitem = menuitem
        self.kwargs = kwargs
        self.progressbar = None
        self.bar_name = menuitem.bar_name
    def get_name(self):
        return self.bar_name
        

    def getProgressBar(self, type):
        # Returns a progress bar object, with no representation.
        if self.progressbar is None:
            self.progressbar = progressbar.ProgressObject(type)
        return self.progressbar

    # "Unthreadable" really means "Run on the same thread as the GUI,
    # and block until finished".  This code actually puts the callback
    # on the GUI thread no matter what.  If threading is not enabled,
    # then the mainthread.mainthread() will return 1, and even if it
    # doesn't, in that case mainthread.runBlock just runs it on the
    # calling thread, which in the unthreaded case is also the main
    # thread. 
    def start(self):
        if self.menuitem.callback is not None:
            if mainthread.mainthread():
                func = self.menuitem.callback
                func_args = (self,)
                func_kwargs = self.kwargs
            else:
                func = mainthread.runBlock
                func_args = (self.menuitem.callback, (self,), self.kwargs)
                func_kwargs = {}
                
            try:
                self.initialize()
                try:
                    func(*func_args, **func_kwargs)
                    # self.menuitem.callback(self, **self.kwargs)
                except SystemExit:
                    raise
                except Exception, exception:
                    if propagate_exceptions:
                        raise
                    else:
                        reporter.error(exception)
            finally:
                self.finalize()
                
    def __getattr__(self, attrname):
        return getattr(self.menuitem, attrname)

    def initialize(self):
        pass
    def finalize(self):
        pass

################
    
class ThreadedWorkerCore(threading.Thread):
    def __init__(self, menuitem, kwargs):
        self.exception_data=None
        self.menuitem = menuitem
        self.kwargs = kwargs
        self.progressbar = None
        self.done = None
        self.bar_name = menuitem.bar_name
        self.threadstate = None
        threading.Thread.__init__(self)
        
    def get_name(self):
        return self.bar_name
    
    def getProgressBar(self, bar_type ="activity"):
        ## should be called within a child thread
        if self.progressbar is not None:
            return self.progressbar
        self.progressbar = progressbar.ProgressObject(bar_type, self.bar_name)
        self.progressbar.registerWorker(self)
        return self.progressbar

    def set_finished_signal(self):
        self.done = 1

    def get_finished_signal(self):
        return self.done

    def cleanupProgressBar(self):
        if self.progressbar:
            self.progressbar.cleanup()
            ## removes itself from list of active progress bars
            
    def cleanupThreadList(self):
        threadstate.cleanList()
        
    def run(self):
        cpbar = progressbar.getProgress()
        self.threadstate = cpbar.get_threadstate()
        try:
            self.initialize()
            try:
                if self.menuitem.callback is not None:
                    self.menuitem.callback(self, **self.kwargs)
            except Exception, exception:
                cpbar.set_failure()
                if propagate_exceptions:
                    self.exception_data=sys.exc_info()
                    sys.exc_clear()
                else:
                    reporter.error(exception)
        finally:
            self.finalize()
            self.set_finished_signal()
            cpbar.make_defunct()
            self.cleanupProgressBar()
            threadmanager.threadManager.removeWorker(self)

        
    def __getattr__(self, attrname):
        return getattr(self.menuitem, attrname)    

    def initialize(self):
        pass
    def finalize(self):
        pass
    
    def cancel(self):
        if self.threadstate:
            threadstate.cancelThread(self.threadstate)
        else:
            print "No threadstate available"


class ThreadedWorker:
    def __init__(self, menuitem, kwargs):
        self.worker = ThreadedWorkerCore(menuitem, kwargs)
        threadmanager.threadManager.newWorker(self.worker)
    def start(self):
        self.worker.cleanupThreadList()
        self.worker.start()
        self.meanwhile()                # back in the main thread...

    def meanwhile(self):  # Default, should be overridden if nontrivial
        pass              # behavior is needed.
    
    def join(self):                     # waits for thread to finish
        self.worker.join()


class TextThreadedWorker(ThreadedWorker):
    def __init__(self, menuitem, kwargs):
        ThreadedWorker.__init__(self, menuitem, kwargs)
        self.initialized = None
        
    def meanwhile(self):
        ## This function should only be called if in text mode
        ## and from the main thread.
        prgrsbar = None
        try:
            thread_running = 1
            while thread_running :
                time.sleep(0.3)
                if self.worker.progressbar is not None:
                    self.progressbar = self.worker.progressbar
                    if  self.initialized is None:
                        ## set progress bar representation and size only once
                        prgrsbar = progressbar.set_ProgressBarRepresentation(self.worker.progressbar) ## for text progress bars
                        prgrsbar.set_window_size(utils.screenwidth())
                        self.initialized = 1 
                    _draw_progressbar(prgrsbar)
                    if prgrsbar.query_stop() or prgrsbar.get_success():
                        thread_running = 0
                        _draw_progressbar(prgrsbar)
                else :
                    ## if progress bar never-ever set then exit loop
                    ## when worker finishes
                    if self.worker.get_finished_signal() is not None:
                        thread_running = 0
                
            
        except KeyboardInterrupt:
            if prgrsbar is not None:
                prgrsbar.stop_it() ## asks for the thread to stop
               ## print "process interrupted"
                _draw_progressbar(prgrsbar)
            else:
                pass ## what to do if there is no progress bar?
        ## reporter.end_progressbar()
        self.join() ## wait for threads to finish before OOF-ing some more.
        ## print "Just hit ctrl-C! in main"

        # For TextThreaded Workers, check if the core has caught an
        # exception and placed it in exception_data.  If so, re-raise
        # it so it gets passed through to the caller of the menu item.
        if propagate_exceptions:
            if self.worker.exception_data is not None:
                raise self.worker.exception_data[0], \
                      self.worker.exception_data[1], \
                      self.worker.exception_data[2]
            

###############################

def _draw_progressbar(pbar):
    reporter.progressbar(pbar.buildprogressbar())
        

###############################
mpiLock = lock.Lock()

## this lock is necessary because MPI does not
## support more than one subthread at a
## time, without running the risk of dead-locking. In order
## to avoid this potential deadlocking
## problem, every parallel worker
## acquires a lock and releases it upon
## completion of the assigned task.
## the MPI lock will not be necessary
## anymore when a fully threadable
## implementation of MPI becomes
## available (MPI-2?).

##################################


# ParallelWorker classes are not instanced unless you're in parallel
# mode, so they are free to use socket2me in this file.

class ParallelThreadedWorkerCore (ThreadedWorkerCore):
    def __init__(self, menuitem, kwargs):
        ThreadedWorkerCore.__init__(self, menuitem, kwargs)
    def initialize(self):
        global mpiLock
        mpiLock.acquire()
    def finalize(self):
        global mpiLock
        mpiLock.release()



class ParallelWork:
    ## ParallelWork encapsulates all the start-up
    ## and clean-up functions that the ParallelWorker
    ## subclasses do, such as communicating between
    ## the different machines behind the scenes
    ## through the adequate (socket) machinery
    def __init__(self, menuitem, kwargs):
        pipeToSocket(menuitem, kwargs)

        ## front-end tells back-end to execute
        ## same menuitem. self.callback() sorts what
        ## is the task that should be performed
        ## by each processor.
        ## back-end bypasses this function

        
class ParallelWorker (NonThreadedWorker, ParallelWork):
    ## Unthreaded Parallel Worker
    def __init__(self,  menuitem, kwargs):
        NonThreadedWorker.__init__(self, menuitem, kwargs)
        ParallelWork.__init__(self, menuitem, kwargs)
        
    
        
class ParallelThreadedWorker (ThreadedWorker, ParallelWork):
    ## Threaded Parallel Worker
    def __init__(self,  menuitem, kwargs):
        self.worker = ParallelThreadedWorkerCore(menuitem, kwargs)
        threadmanager.threadManager.newWorker(self.worker)
        ParallelWork.__init__(self, menuitem, kwargs)
        
    def meanwhile(self):
        pass ## back-end should always pass and give control
    ## to socketinput, after launching the job
    
class ParallelTextThreadedWorker(TextThreadedWorker, ParallelWork):
    ## Text mode Threaded Parallel Worker
    ## sould only be called by front end
    def __init__(self,  menuitem, kwargs):
        self.worker = ParallelThreadedWorkerCore(menuitem, kwargs)
        threadmanager.threadManager.newWorker(self.worker)
        self.initialized = None
        ParallelWork.__init__(self, menuitem, kwargs)
        
        
###############################
# getThreadedWorker() is redefined if the GUI is loaded in workerGUI.py

def getThreadedWorker(menuitem, argdict):
    ## serial workers
    return TextThreadedWorker(menuitem, argdict)

def getThreadedParallelWorker(menuitem, argdict):
    ## parallel workers
    try:
        from oof2.SWIG.common import mpitools
        if mpitools.rank() == 0:
            ## front-end runs threaded workers with text progress bars
            return ParallelTextThreadedWorker(menuitem, argdict)
        ## back-end runs a threaded worker too, but no text progress bar
        ## is available. Back-end, instead, runs a thread, and the
        ## "front-end' of the back-end waits on the socket parser
        return ParallelThreadedWorker(menuitem, argdict)
    except ImportError:
            raise
    
