# -*- python -*-
# $RCSfile: orientationmatrix.py,v $
# $Revision: 1.27 $
# $Author: langer $
# $Date: 2005/05/13 18:44:10 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Orientation is another convertible registered type, similar
# to isocijkl.  Orientation can be expressed in terms of various
# angle conventions, all of which get converted to the base
# type, which is a SWIG'd EulerAngle object.

from oof2.common import registeredclass
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump
from oof2.SWIG.engine.eulerangle import EulerAngle
import math

FloatParameter = parameter.FloatParameter

# Base class.  Should be able to answer questions about its values
# in different representations, using the member *instance*'s to_base
# functions.  
class Orientation(registeredclass.ConvertibleRegisteredClass):
    registry = []
    # Orientation needs a special equality checker, because
    # it's value is not a composite Python object, as assumed by
    # RegisteredClass.
    def __eq__(self,other):
        if other is None:
            return 0
        if self.__class__!=other.__class__:
            return 0
        self_base = self.to_base()
        other_base = other.to_base()
        if self_base.alpha()!=other_base.alpha() or \
           self_base.beta()!=other_base.beta() or \
           self_base.gamma()!=other_base.gamma():
            return 0
        return 1
    def __ne__(self,other):
        return 1-self.__eq__(other)

    tip = "The orientation of a three dimensional object."

    discussion = """<para>
    There are many different ways of describing the orientation of a
    three dimensional object.  <application>OOF1</application> used
    only Euler Angles, assuming that everybody used Euler angles, but
    it turns out that not only does everybody not use Euler angles,
    but there are differences of opinion about how Euler angles should
    be defined.  So &oof2; allows you to choose among many different
    ways of defining orientations, some of which are Euler angles, and
    none of which are called Euler angles.  The <link
    linkend='RegisteredClass:Abg'><classname>Abg</classname></link> is
    the one used in <application>OOF1</application>.
    </para>"""
        

# The base value type is EulerAngle, imported from oof2.SWIG.engine.eulerangle.


# Registration object hosts the "call points" for to_base
# routines.  These are used by the GUI when switching types.
class OrientationRegistration(registeredclass.ConvertibleRegistration):
    def __init__(self, name, subclass, ordering,
                 from_base, to_base, params=[], tip=None, discussion=None):
        registeredclass.Registration.__init__(self, name,
                                              Orientation, subclass,
                                              ordering, from_base=from_base,
                                              to_base=to_base, params=params,
                                              tip=tip, discussion=discussion)

#########################################################################
#########################################################################

# Conversion for the symmetric matrix representation of the Cijkl
# matrix.  Same internal data representation as the base value type.

class Abg(Orientation):
    def __init__(self, alpha, beta, gamma):
        # Set the parameters in the registration object.  This is
        # redundant (but harmless) when the object is instanced from
        # the registration.
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

def _abg_to_base(reg, vlist = None):
    vset = vlist or reg.getParamValues()
    return EulerAngle(*[math.pi/180.*deg for deg in vset])

def _abg_from_base(reg, base):
    if isinstance(base, EulerAngle):
        reg.setDefaultParams([180./math.pi*rad for rad in (base.alpha(),
                                                           base.beta(),
                                                           base.gamma())])

OrientationRegistration(
    'Abg',
    Abg,
    1,
    from_base=_abg_from_base,
    to_base=_abg_to_base,
    params=[FloatParameter('alpha',0.0, ## tip='c-axis tilt'),
                           tip='a rotation of the z-axis about the y-axis'),
            FloatParameter('beta', 0.0, ##  tip='c-axis rotation'),
                           tip='a rotation of the z-axis about itself'),
            FloatParameter('gamma', 0.0, ## tip='z-axis rotation')],
                           tip='a rotation of the z-axis about itself')],
    tip='Euler angles (alpha, beta, gamma) are applied: first beta, then alpha, and finally gamma',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/abg.xml')
    )

#######################################
#######################################

# Utility function for converting a rotation matrix (stored as a
# Python dictionary indexed by tuples) into an EulerAngle object.
# Used by most of the other conversion routines, since the matrix
# is a more convenient base-like type.
#   The reason for not using the matrix itself as the base type
# is that we want to keep all the conversion code in this file,
# if possible.
def _matrix_to_base(matrix):
    cosa = matrix[(2,2)]
    sina = math.sqrt(matrix[(2,0)]*matrix[(2,0)]+
                     matrix[(2,1)]*matrix[(2,1)])
    #
    # sina might be zero.
    if sina != 0.0:
        cosb = matrix[(2,0)]/sina
        sinb = matrix[(2,1)]/sina
        #
        cosg = -matrix[(0,2)]/sina
        sing = matrix[(1,2)]/sina
        #
    else:
        if cosa > 0.0: # i.e., it's 1.0, but there's roundoff.
            cosb = matrix[(0,0)]
            sinb = matrix[(0,1)]
            sing = 0.0
            cosg = 1.0
        else: # In this case, cosa must be -1.
            cosb = -matrix[(0,0)]
            sinb = -matrix[(0,1)]
            sing = 0.0
            cosg = 1.0
   
    return EulerAngle(math.atan2(sina,cosa),
                      math.atan2(sinb,cosb),
                      math.atan2(sing,cosg))

##    abg = [ v*180.0/math.pi for v in [ math.atan2(sina,cosa),
##                                       math.atan2(sinb,cosb),
##                                       math.atan2(sing,cosg) ] ]

##    return apply(EulerAngle, abg)


def _base_to_matrix(base):
    [[sina,cosa],[sinb,cosb],[sing,cosg]] = \
             [ [ math.sin(angle), math.cos(angle) ]  for angle in
##                    [ degrees*math.pi/180.0 for degrees in
                      [base.alpha(), base.beta(), base.gamma()]
##                      ]
               ]

    matrix = {}
    
    matrix[(0,0)] = cosa*cosb*cosg - sinb*sing
    matrix[(0,1)] = cosa*sinb*cosg + cosb*sing
    matrix[(0,2)] = -sina*cosg
    #
    matrix[(1,0)] = -cosa*cosb*sing - sinb*cosg
    matrix[(1,1)] = -cosa*sinb*sing + cosb*cosg
    matrix[(1,2)] = sina*sing
    #
    matrix[(2,0)] = sina*cosb
    matrix[(2,1)] = sina*sinb
    matrix[(2,2)] = cosa
    
    return matrix


    

###########################################################################
###########################################################################

# Goldstein's "X" convention.  This may have some other more
# descriptive name, but I don't know what it is.
class X(Orientation):
    def __init__(self, phi, theta, psi):
        self.phi = phi
        self.theta = theta
        self.psi = psi
##    def __repr__(self):
##        return "X(phi=%f, theta=%f, psi=%f)" % \
##               (self.phi, self.theta, self.psi)

# vlist should again be an indexable with vlist[0]=phi, vlist[1]=theta,
# vlist[2]=phi.  Converts by way of the matrix.
def _x_to_base(x_reg, vlist=None):
    (phi, theta, psi) = vlist or x_reg.getParamValues()

    [[sinph,cosph],[sinth,costh],[sinps,cosps]] = \
       [ [ math.sin(angle), math.cos(angle) ] for angle in
         [ degrees*math.pi/180.0 for degrees in (phi, theta, psi) ] ]

    data = {}
    
    data[(0,0)] = cosps*cosph - costh*sinph*sinps
    data[(0,1)] = cosps*sinph + costh*cosph*sinps
    data[(0,2)] = sinps*sinth
    #
    data[(1,0)] = -sinps*cosph - costh*sinph*cosps
    data[(1,1)] = -sinps*sinph + costh*cosph*cosps
    data[(1,2)] = cosps*sinth
    #
    data[(2,0)] = sinth*sinph
    data[(2,1)] = -sinth*cosph
    data[(2,2)] = costh
    
    return _matrix_to_base(data)
       


def _x_from_base(x_reg, base):
    if isinstance(base, EulerAngle):
        data = _base_to_matrix(base)
        
        costh = data[(2,2)]
        sinth = math.sqrt(data[(0,2)]*data[(0,2)]+
                          data[(1,2)]*data[(1,2)])
        #
        if sinth != 0.0:
            cosps = data[(1,2)]/sinth
            sinps = data[(0,2)]/sinth
            #
            cosph = -data[(2,1)]/sinth
            sinph = data[(2,0)]/sinth
            #
        else:
            if costh > 0.0: # Again, since sinth=0, this means costh=1.0
                cosph = data[(0,0)]
                sinph = data[(0,1)]
                cosps = 1.0
                sinps = 0.0
            else: # costh is -1.
                cosph = -data[(0,0)]
                sinph = data[(0,1)]
                cosps = 1.0
                sinph = 0.0

        x_reg.setDefaultParams(
            [ v*180.0/math.pi for v in [math.atan2(sinph,cosph),
                                        math.atan2(sinth,costh),
                                        math.atan2(sinps,cosps)] ] )

OrientationRegistration(
    'X',
    X,
    2,
    from_base=_x_from_base,
    to_base=_x_to_base,
    params=[FloatParameter('phi', 0.0),
            FloatParameter('theta', 0.0),
            FloatParameter('psi', 0.0)],
    tip="Goldstein's X convention for 3D orientations.")


###########################################################################
###########################################################################

# The "aerodynamic" XYZ convention, with each rotation about a different
# principal axis.  Again the name is from Goldstein.

class XYZ(Orientation):
    def __init__(self, phi, theta, psi):
        self.phi = phi
        self.theta = theta
        self.psi = psi
##    def __repr__(self):
##        return "XYZ(phi=%f, theta=%f, psi=%f)" % \
##               (self.phi, self.theta, self.psi)

# vlist should again be an indexable with vlist[0]=phi, vlist[1]=theta,
# vlist[2]=phi.  Converts by way of the matrix.
def _xyz_to_base(xyz_reg, vlist=None):
    (phi, theta, psi) = vlist or xyz_reg.getParamValues()

    [[sinph,cosph],[sinth,costh],[sinps,cosps]] = \
       [ [ math.sin(angle), math.cos(angle) ] for angle in
         [ degrees*math.pi/180.0 for degrees in (phi, theta, psi) ] ]

    data = {}
    
    data[(0,0)] = costh*cosph
    data[(0,1)] = costh*sinph
    data[(0,2)] = -sinth
    #
    data[(1,0)] = sinps*sinth*cosph - cosps*sinph
    data[(1,1)] = sinps*sinth*sinph + cosps*cosph
    data[(1,2)] = costh*sinps
    #
    data[(2,0)] = cosps*sinth*cosph + sinps*sinph
    data[(2,1)] = cosps*sinth*sinph - sinps*cosph
    data[(2,2)] = costh*cosps
    
    return _matrix_to_base(data)
       


def _xyz_from_base(xyz_reg, base):
    if isinstance(base, EulerAngle):
        data = _base_to_matrix(base)
        
    sinth = -data[(0,2)]
    costh = math.sqrt(data[(0,0)]*data[(0,0)]+
                      data[(0,1)]*data[(0,1)])
    #
    if costh!=0.0:
        sinph = data[(0,1)]/costh
        cosph = data[(0,0)]/costh
        #
        sinps = data[(1,2)]/costh
        cosps = data[(2,2)]/costh
        #
    else:
        if sinth > 0.0: # i.e. if sinth is 1.0, theta is pi/2.
            sinph = data[(1,0)]
            cosph = data[(1,1)]
            sinps = 0.0
            cosps = 1.0
        else: # sinth < 0, theta is -pi/2.
            sinph = -data[(1,0)]
            cosph = data[(1,1)]
            sinps = 0.0
            cosps = 1.0

    xyz_reg.setDefaultParams(
        [ v*180.0/math.pi for v in [math.atan2(sinph,cosph),
                                    math.atan2(sinth,costh),
                                    math.atan2(sinps,cosps)] ] )


OrientationRegistration(
    'XYZ', XYZ, 3,
    from_base=_xyz_from_base,
    to_base=_xyz_to_base,
    params=[FloatParameter('phi', 0.0),
            FloatParameter('theta', 0.0),
            FloatParameter('psi', 0.0)],
    tip='The "aerodynamic" XYZ convention for specifying 3D orientations.')


###########################################################################
###########################################################################

# Quaternions, for two reasons:  Firstly, it's fun.  Secondly,
# if you're going to have the axis-and-angle parametrization, you've
# already done most of the work for the quaternions.

# One result of this is that there is substantial, informally-shared
# structure between the Quaternion and Axis representations.  This
# is the sort of thing that would normally be done by inheritance, but
# isn't, because the "to_base" and "from_base" function are not class
# members, but rather are data in the registration instances.
#   Be aware that _quaternion_to_base and _quaternion_from_matrix
# are also called by _axis_to_base and _axis_from_base (see below). 

class Quaternion(Orientation):
    def __init__(self, e0, e1, e2, e3):
        self.e0 = e0
        self.e1 = e1
        self.e2 = e2
        self.e3 = e3
##    def __repr__(self):
##        return "Quaternion(e0=%f, e1=%f, e2=%f, e3=%f)" % \
##               (self.e0, self.e1, self.e2, self.e3)

# vlist should again be an indexable with vlist[0]=e0, vlist[1]=e1,
# vlist[2]=e2, vlist[3]=e3.
def _quaternion_to_base(quat_reg, vlist=None):
    (e0, e1, e2, e3) = vlist or quat_reg.getParamValues()

    # Un-normalized quaternions are probably dangerous, they
    # result in all the entries in the matrix being too large by a
    # factor of the normalization squared, which means the deduced
    # cosine from the "corner" entry will be wrong.  So, normalize.
    norm2 = e0*e0+e1*e1+e2*e2+e3*e3

    if norm2==0.0: # Presume the user meant "no rotation".
        e0 = 1.0
        e1 = 0.0
        e2 = 0.0
        e3 = 0.0
    else:
        e0 /= math.sqrt(norm2)
        e1 /= math.sqrt(norm2)
        e2 /= math.sqrt(norm2)
        e3 /= math.sqrt(norm2)
    
    matrix = {}
        
    matrix[(0,0)] = e0*e0 + e1*e1 - e2*e2 - e3*e3
    matrix[(0,1)] = 2.0*(e1*e2+e0*e3)
    matrix[(0,2)] = 2.0*(e1*e3-e0*e2)
    #
    matrix[(1,0)] = 2.0*(e1*e2-e0*e3)
    matrix[(1,1)] = e0*e0-e1*e1 + e2*e2-e3*e3
    matrix[(1,2)] = 2.0*(e2*e3+e0*e1)
    #
    matrix[(2,0)] = 2.0*(e1*e3+e0*e2)
    matrix[(2,1)] = 2.0*(e2*e3-e0*e1)
    matrix[(2,2)] = e0*e0-e1*e1 - e2*e2+e3*e3
        
    return _matrix_to_base(matrix)
       

# Make a special, lower-level function which the axis-and-angle
# representation can borrow.
def _quaternion_from_matrix(matrix):
    e0e1 = (matrix[(1,2)]-matrix[(2,1)])/4.0
    e0e2 = (matrix[(2,0)]-matrix[(0,2)])/4.0
    e0e3 = (matrix[(0,1)]-matrix[(1,0)])/4.0
    #
    e1e2 = (matrix[(0,1)]+matrix[(1,0)])/4.0
    e1e3 = (matrix[(0,2)]+matrix[(2,0)])/4.0
    e2e3 = (matrix[(1,2)]+matrix[(2,1)])/4.0
    #
    e0e0 = (matrix[(0,0)]+matrix[(1,1)]+matrix[(2,2)]+1.0)/4.0
    #
    # The unstated fourth equation is the normalization
    # condition, e0^2+e1^2+e2^2+e3^2 == 1.
    # Possible cases: if e1, e2, and e3 are all zero, then
    # sin(phi/2) has to be zero, and if the angle is zero
    # then the axis doesn't matter.
    
    # If e0 is zero, then cos(phi/2) is 0, phi is +/- pi.
    # sin(phi/2) = 1, and there's an ambiguity in the
    # axis direction - 180 degree rotations about plus or minus
    # axis is of course the same..
    if e0e0==0.0:
        e1e1 = -(matrix[(1,1)]+matrix[(2,2)])/2.0
        e2e2 = -(matrix[(0,0)]+matrix[(2,2)])/2.0
        e3e3 = -(matrix[(0,0)]+matrix[(1,1)])/2.0
        #
        e0 = 0.0
        if e3e3 != 0.0:
            e3 = math.sqrt(e3e3)
            e2 = e2e3/e3
            e1 = e1e3/e3
        elif e2e2 != 0.0:
            e2 = math.sqrt(e2e2)
            e3 = e2e3/e2
            e1 = e1e2/e2
        else: # If e0 is zero, one of these has to be nonzero.
            e1 = math.sqrt(e1e1)
            e3 = e1e3/e1
            e2 = e1e2/e1
    # If e0^2 is nonzero, we can assume a positive e0, restricting
    # the angle phi to -pi -> pi.
    else:
        e0 = math.sqrt(e0e0)
        e1 = e0e1/e0
        e2 = e0e2/e0
        e3 = e0e3/e0
        
    return [e0,e1,e2,e3]


# Then from_base is a simple wrapper for this function.
def _quaternion_from_base(quat_reg, base):
    if isinstance(base, EulerAngle):
        matrix = _base_to_matrix(base)

        qlist = _quaternion_from_matrix(matrix)
        
        quat_reg.setDefaultParams(qlist)


OrientationRegistration(
    'Quaternion',
    Quaternion,
    5,
    from_base=_quaternion_from_base,
    to_base=_quaternion_to_base,
    params=[FloatParameter('e0', 0.0),
            FloatParameter('e1', 0.0),
            FloatParameter('e2', 0.0),
            FloatParameter('e3', 0.0)],
    tip="The Quaternion representation for 3D orientations.")

###########################################################################
###########################################################################

# Axis-and-angle.  Also uses _quaternion_to_base_ and
# _quaternion_from_base, but since they're data, this isn't
# proper (structural) inheritance.


class Axis(Orientation):
    def __init__(self, angle, x, y, z):
        self.angle = angle
        self.x = x
        self.y = y
        self.z = z
##    def __repr__(self):
##        return "Axis(angle=%f, x=%f, y=%f, z=%f)" % \
##               (self.angle, self.x, self.y, self.z)

# vlist should again be an indexable with vlist[0]=angle, vlist[1]=x,
# vlist[2]=y, vlist[3]=z.
def _axis_to_base(axis_reg, vlist=None):
    (angle, x, y, z) = vlist or axis_reg.getParamValues()

    # Convert to quaternions, it's easy.
    
    # The math is written weirdly to emphasize that the arg is theta/2
    cos_half_theta = math.cos( (math.pi/180.0)*(angle/2.0) )
    sin_half_theta = math.sin( (math.pi/180.0)*(angle/2.0) )
    
    norm = math.sqrt(x*x+y*y+z*z)
    
    quat = [cos_half_theta, 0.0, 0.0, 0.0]
    # Assume z-axis in the absence of clues.
    if norm == 0.0:
        quat[0] = 0.0
        quat[1] = 0.0
        quat[2] = sin_half_theta
    else:
        quat[1] = x*(sin_half_theta/norm)
        quat[2] = y*(sin_half_theta/norm)
        quat[3] = z*(sin_half_theta/norm)

    return _quaternion_to_base(None,quat)


def _axis_from_base(axis_reg, base):
    if isinstance(base, EulerAngle):
        matrix = _base_to_matrix(base)

        quat = _quaternion_from_matrix(matrix)

        sin_half_theta = math.sqrt(quat[1]*quat[1]+quat[2]*quat[2]+
                                   quat[3]*quat[3])
        cos_half_theta = math.sqrt(quat[0]*quat[0])

        theta = (180.0/math.pi)*2.0*math.atan2(sin_half_theta,cos_half_theta)
        

        axis = [theta, 0.0, 0.0, 0.0]
        
        # Assume z-axis in the absence of clues.
        if sin_half_theta == 0.0:
            axis[1] = 0.0
            axis[2] = 0.0
            axis[3] = 1.0
        else:
            axis[1] = quat[1]/sin_half_theta
            axis[2] = quat[2]/sin_half_theta
            axis[3] = quat[3]/sin_half_theta

        axis_reg.setDefaultParams(axis)


OrientationRegistration(
    'Axis',
    Axis,
    4,
    from_base=_axis_from_base,
    to_base=_axis_to_base,
    params=[FloatParameter('angle', 0.0, tip='Rotation angle.'),
            FloatParameter('x', 0.0, tip='x component of rotation axis.'),
            FloatParameter('y', 0.0, tip='y component of rotation axis.'),
            FloatParameter('z', 0.0, tip='z component of rotation axis.')],
    tip="Orientation specified by rotation axis and angle.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/axis.xml')
    )




###########################################################################
###########################################################################

## Rodrigues vector. Another way of describing crystal orientations.
## This form is quite popular in the texture community; it is particularly
## useful to describe fiber-texture and poling in ferroelectrics. --REG
class Rodrigues(Orientation):
    def __init__(self, r1, r2, r3):
        self.r1 = r1
        self.r2 = r2
        self.r3 = r3


def _rodrigues_to_base(r_reg, vlist=None):
    (r1, r2, r3) = vlist or r_reg.getParamValues()
    
    magnitude_of_r = r1*r1+r2*r2+r3*r3 ## This is the magnitude square of the Rodrigues vector
    qo = 1.0/math.sqrt(1.0 + magnitude_of_r)
    q1 = qo*r1
    q2 = qo*r2
    q3 = qo*r3
    local_qlist = [qo, q1, q2, q3]
    ## The above equations were taken from
    ## A. Morawiec, and D. P. Field "Rodrigues parametrization for
    ## orientation and misorientation distributions". Phylosophical
    ## Magazine A, 1996, Vol. 73, No. 4, 1113-1130
    return _quaternion_to_base(None, local_qlist)
       


def _rodrigues_from_base(r_reg, base):
    if isinstance(base, EulerAngle):
        matrix = _base_to_matrix(base)
    quat = _quaternion_from_matrix(matrix)
    ##rlist = [0.0,0.0,0.0]
    if quat[0] != 0 :
        rlist = [quat[1]/quat[0], quat[2]/quat[0], quat[3]/quat[0]]
        ## The above relationship was extracted from
        ## A. Heinz, and P. Neumann "Representation of Orientation and
        ## Disorientation Data for Cubic, Hexagonal, Tetragonal, and
        ## Orthorhombic Crystals". Acta Cryst. (1991) A47, 780-789
    else:
        ## Rodrigues vector has a degeneracy point at theta = \pm PI
        ## In order to --remove it--, I mapped it back to a twist of
        ## PI around whatever value the normal vector points to
        quat_normal = math.sqrt(quat[1]*quat[1]+quat[2]*quat[2]+quat[3]*quat[3])
        if quat_normal != 0:
            quat[1] /= quat_normal ## Here, we use quat to create
            quat[2] /= quat_normal ## a local normal vector
            quat[3] /= quat_normal
        rlist = [2.0e9*quat[1], 2.0e9*quat[2],2.0e9*quat[3]]
        ## Here, 2.0e9 corresponds to the tangent of PI-1e-9 radians,
        ## which is a "reasonable" approximation to PI...
    r_reg.setDefaultParams(rlist)

OrientationRegistration(
    'Rodrigues',
    Rodrigues,
    6,
    from_base=_rodrigues_from_base,
    to_base=_rodrigues_to_base,
    params=[FloatParameter('r1', 0.0),
            FloatParameter('r2', 0.0),
            FloatParameter('r3', 0.0)],
    tip='Rodrigues vector representation for 3D orientations.')

