# -*- python -*-
# $RCSfile: outputClones.py,v $
# $Revision: 1.62 $
# $Author: langer $
# $Date: 2005/05/13 18:44:10 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

# A smattering of pre-defined outputs for the convenience of OOF users.

from oof2.SWIG.common import coord
from oof2.SWIG.common import ooferror
from oof2.SWIG.engine import field
from oof2.SWIG.engine import flux
from oof2.SWIG.engine import invariant
from oof2.SWIG.engine import outputval
from oof2.common import debug
from oof2.common import enum
from oof2.common import primitives
from oof2.common.IO import output
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump
from oof2.engine.IO import meshparameters
from types import *
import math

IndexableTypeChecker = parameter.IndexableTypeChecker
Input = output.Input
Output = output.Output
FloatParameter = parameter.FloatParameter


# Examples of Outputs

# The 'func' of an output should have the following form:
#    def func([input1, ...,][mesh,][elements,][coords,][parameters])
# where the inputs will be passed as lists in order at the beginning
# of the argument list, and the rest of the arguments may be in any
# order. If any of mesh,elements,coords,parameters (or all) are not
# desired, then they can all be slurped into **kwargs and discarded
# (as in the first example below).

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

def _pos(mesh,elements,coords,*args,**kwargs):
    # The argument "elements" is a list of Elements or Edges, possibly
    # mixed together. The argument "coords" is a list of lists.  Each
    # Element in the elements list corresponds to a list of
    # MasterCoords in the coords list, and each Edge corresponds to a
    # list of doubles (ie, master coords for the Edge, in the range
    # [0,1)).
    ans = []
    for element,coordList in zip(elements,coords):
        ans += element.position(coordList)
    return ans

posOutput = Output(func=_pos,name="original position",
                   tip="Original positions of points in physical space.",
                   oType = (primitives.Point, coord.Coord))

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

class SpaceComponent(enum.EnumClass('x', 'y', 'z')):
    tip='Components of vectors.'
    discussion="""<para>
    <classname>SpaceComponent</classname> is used by various
    <link
    linkend='Section:Output'><classname>Outputs</classname></link> to
    choose one of the components of a vector quantity.
    </para>"""
    
class InPlaneSpaceComponent(enum.EnumClass('x', 'y')):
    tip="The in-plane components of vectors."
    discussion="""<para>
    <classname>InPlaneSpaceComponent</classname> is used by various
    <link
    linkend='Section:Output'><classname>Outputs</classname></link> to
    choose one of the in-plane components of a vector quantity.
    </para>"""

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

# FieldOutput is for fields which are defined on the Mesh.
# FieldAnywayOutput is for fields, even if they're not defined on the
# mesh.

def _field(mesh, elements, coords, parameters, *args, **kwargs):
    feeled = parameters[0].value
    ## TODO: rewrite in terms of list comprehension [x for x in list]
    ## or simply preallocate space in ans.  Appending to the list is
    ## slow.  Fix this in all output funcs, including
    ## propertyoutput_.py.
    ans = []
    for element, ecoords in zip(elements, coords):
        ans += element.outputFields(mesh, feeled, ecoords)
    return ans

FieldOutput = Output(
    name = "field",
    func = _field,
    oType = outputval.OutputValPtr,
    parameters = meshparameters.FieldParameter("field", outofplane=1,
                                               tip=parameter.emptyTipString),
    tip="Compute Field values.",
    discussion='<para>Compute the value of the given &field; on a &mesh;.</para>')

def _fieldAnyway(mesh, elements, coords, parameters, *args, **kwargs):
    feeled = parameters[0].value
    ans = []
    for element, ecoords in zip(elements, coords):
        ans += element.outputFieldsAnyway(mesh, feeled, ecoords)
    return ans

FieldAnywayOutput = Output(
    name = "field anyway",
    func=_fieldAnyway,
    oType=outputval.OutputValPtr,
    parameters = meshparameters.FieldAnywayParameter("field", outofplane=1))

############

def _fieldderiv(mesh, elements, coords, parameters, *args, **kwargs):
    feeled = parameters[0].value
    dcomp = parameters[1].value
    ans = []
    for element, ecoords in zip(elements, coords):
        ans += element.outputFieldDerivs(mesh, feeled, dcomp, ecoords)
    return ans

FieldDerivOutput = Output(
    name = "field derivative",
    func = _fieldderiv,
    oType = outputval.OutputValPtr,
    parameters = [meshparameters.FieldParameter("field", outofplane=1,
                                                tip=parameter.emptyTipString),
                  enum.EnumParameter("derivative", InPlaneSpaceComponent,
                                     tip='Which derivative to take.')],
    tip='Compute derivatives of Fields.',
    discussion='<para>Compute the spatial derivative of a &field; on a &mesh;.</para>')

############

def _flux(mesh, elements, coords, parameters, *args, **kwargs):
    flucks = parameters[0].value
    ans = []
    for element, ecoords in zip(elements, coords):
        element.begin_material_computation(mesh)
        ans += element.outputFluxes(mesh, flucks, ecoords)
        element.end_material_computation(mesh)
    return ans

FluxOutput = Output(
    name = "flux",
    func = _flux,
    oType = outputval.OutputValPtr,
    parameters = meshparameters.FluxParameter("flux",
                                              tip=parameter.emptyTipString),
    tip='Compute Flux values.',
    discussion='<para>Compute the value of the given &flux; on a &mesh;.</para>')

################

def _component(fieldvals, parameters, *args, **kwargs):
    if fieldvals:
        compstr = parameters[0].value   # a string, "x" or "xy" or the like
        comp = fieldvals[0].getIndex(compstr) # convert string to IndexP
        return [f[comp] for f in fieldvals]
    return []
        
ComponentOutput = Output(
    name = "component",
    func = _component,
    oType = FloatType,
    inputs = Input('field', outputval.OutputValPtr),
    parameters = meshparameters.FieldIndexParameter('component',
                                                    tip='Which component to take.'))

FieldCompOutput = ComponentOutput.clone(
    name="field component",
    inputs=Input("field", outputval.OutputValPtr, FieldOutput.clone()),
    tip='Compute a component of a Field.',
    discussion=
    '<para>Compute a single component of a &field; on a &mesh;.</para>')

FieldDerivCompOutput = ComponentOutput.clone(
    name="field derivative component",
    inputs=Input("field", outputval.OutputValPtr, FieldDerivOutput.clone()),
    tip='Compute a component of a Field derivative.',
    discussion='<para>Compute a single component of a &field; derivative on a &mesh;.</para>')

FluxCompOutput = ComponentOutput.clone(
    name="flux component",
    inputs=Input("field", outputval.OutputValPtr, FluxOutput.clone()),
    tip='Compute a component of a Flux.',
    discussion=
    '<para>Compute a single component of a &flux; on a &mesh;.</para>')

###############

def _invariant(fieldvals, parameters, *args, **kwargs):
    inv = parameters[0].value
    return map(inv, fieldvals)

InvariantOutput = Output(
    name="invariant",
    func=_invariant,
    oType=FloatType,
    inputs=Input('field', outputval.OutputValPtr),
    parameters=invariant.InvariantParameter("invariant",
                                            tip=parameter.emptyTipString))

FieldInvariantOutput = InvariantOutput.clone(
    name="field invariant",
    inputs=Input("field", outputval.OutputValPtr, FieldOutput.clone()),
    tip='Compute invariants of Fields.',
    discussion='<para>Compute invariants of a &field; on a &mesh;.</para>')

FluxInvariantOutput = InvariantOutput.clone(
    name="flux invariant",
    inputs=Input("field", outputval.OutputValPtr, FluxOutput.clone()),
    tip='Compute invariants of Fluxes.',
    discussion='<para>Compute invariants of a &flux; on a &mesh;.</para>')

FieldDerivInvariantOutput = InvariantOutput.clone(
    name="field derivative invariant",
    inputs=Input("field", outputval.OutputValPtr, FieldDerivOutput.clone()),
    tip='Compute invariants of Field derivatives.',
    discussion='<para>Compute invariants of the derivative of a &field; on a &mesh;</para>')


#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

from oof2.common import strfunction

def _functionOutput(elements, coords, parameters, *args, **kwargs):
    ans = []
    function = parameters[0].value
    for element, coordlist in zip(elements, coords):
        realcoords = [element.from_master(c[0], c[1]) for c in coordlist]
        ans += [function(coord[0], coord[1]) for coord in realcoords]
    return ans

FunctionOutput = Output(
    func=_functionOutput,
    name="Function",
    oType=FloatType,
    parameters=(strfunction.XYStrFunctionParameter("f",
                                                   tip="Function of x and y."),))

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

def _pointSum(first, second, parameters, *args, **kwargs):
    a = parameters[0].value
    b = parameters[1].value
    ans = [a*f+b*s for f,s in zip(first, second)]
    return ans

PointSumOutput = \
               Output(name="point sum", func=_pointSum,
                      oType=primitives.Point,
                      inputs=(Input("point1", (primitives.Point, coord.Coord)),
                              Input("point2", (primitives.Point, coord.Coord))),
                      parameters=(FloatParameter("a", default=1.0),
                                  FloatParameter("b", default=1.0)))

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#


# Scale values to lie between 0 and 1.

def _rescaleOutput(values, parameters, *args, **kwargs):
    targetmin = parameters[0].value
    targetmax = parameters[1].value
    minval = 1.e100
    maxval = -minval
    for vlist in values:
        for v in vlist:
            if v < minval: minval = v
            if v > maxval: maxval = v
    def rescale(x, mn=minval, mx=maxval, tmin=targetmin, tmax=targetmax):
        if mx == mn:
            return 0.5*(tmin + tmax)    # arbitrary
        return tmin + (tmax-tmin)*(x-mn)/(mx-mn)
    ans = []
    for vlist in values:
        ans += [map(rescale, vlist)]
    return ans

RescaleOutput = \
              Output(name="rescale",
                     func=_rescaleOutput,
                     parameters=(FloatParameter("minimum", default=0.),
                                 FloatParameter("maximum", default=1.)),
                     inputs=Input("scalar",FloatType),
                     oType=FloatType)

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

# Multiply by a scalar

def _multOutput(values, parameters, *args, **kwargs):
    factor = parameters[0].value
    return [[factor*x for x in vlist] for vlist in values]

ScalarMultiplyOutput = Output(
    name="scalar multiply",
    func=_multOutput,
    parameters=FloatParameter("factor", default=1.),
    inputs=Input("scalee", outputval.OutputValPtr),
    oType = outputval.OutputValPtr,
    tip="Multiply the input by a constant scalar factor.")

NegateOutput = ScalarMultiplyOutput.clone(
    name="negate", tip="Multiply the input by -1.")
NegateOutput.plugParameter(which="factor", value=-1.0)
