# -*- python -*-
# $RCSfile: bdycondition.py,v $
# $Revision: 1.65 $
# $Author: langer $
# $Date: 2005/08/08 21:39:33 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# A new, leaner, meaner, boundary condition class.  For the
# point-like ones, the iteration over nodes and point evaluation,
# as required, will be done elsewhere.  All we need to be able
# to do here is take a node and a profile result, and apply it.

from oof2.SWIG.common import ooferror
from oof2.SWIG.common import switchboard
from oof2.SWIG.engine import boundarycond
from oof2.common import debug
from oof2.common import labeltree
from oof2.common import registeredclass
from oof2.common.IO import automatic
from oof2.common.IO import parameter
from oof2.common.IO import whoville
from oof2.common.IO import xmlmenudump
from oof2.engine import problem
from oof2.engine import profile
from oof2.engine import skeletoncontext
from oof2.engine.IO import meshparameters
import oof2.engine.mesh


##def bcnameResolver(param, name):
##    if param.automatic():
##        basename = 'bc'
##    else:
##        basename = name
##    meshname = param.group['mesh'].value
##    if meshname is not None:
##        meshpath = labeltree.makePath(meshname)
##        meshcontext = oof2.engine.mesh.meshes[meshpath]
##        return meshcontext.uniqueBCName(basename)



##nameparam = parameter.AutomaticNameParameter(
##    'name', value=automatic.automatic, tip="Select a name.",
##    resolver=bcnameResolver)

##meshparam = whoville.WhoParameter("mesh", oof2.engine.mesh.meshes,
##                                  tip="Specify the mesh.")

##bccontextgroup = parameter.ParameterGroup(
##    nameparam,
##    meshparam)




# It's useful in the editor if all the BC's have shared
# parameters, to the greatest degree possible.

bc_field_param_group = parameter.ParameterGroup(
    meshparameters.FieldParameter('field', tip=parameter.emptyTipString),
    meshparameters.FieldIndexParameter('field_component',
                                       tip=parameter.emptyTipString))

bc_equation_param_group = parameter.ParameterGroup(
    meshparameters.EquationParameter('equation', tip=parameter.emptyTipString),
    meshparameters.FieldIndexParameter('eqn_component',
                                       tip=parameter.emptyTipString))


# bc_boundary_param = skeletongroupparams.BoundaryParameter(
#     "boundary",tip="Boundary to which this condition applies.")
bc_boundary_param = meshparameters.MeshBoundaryParameter(
    "boundary", tip="Boundary to which this condition applies.")

bc_edge_bdy_param = meshparameters.MeshEdgeBdyParameter(
    "boundary", tip="Edge boundary to which this condition applies.")

bc_point_bdy_param = meshparameters.MeshPointBdyParameter(
    "boundary", tip="Point boundary to which this condition applies.")


# TODO: Boundary copy operations could transfer the explicit-enable
# state of the system easily, the constructors all take this argument.
# It's only used by the promotion machinery for making Dirichlet BC's
# out of floating BCs.

# TODO: Float BC's and Dirichlet BCs which collide but are disabled
# should probably not throw exceptions.

# Boundary conditions are constructed in association with a particular
# profile and boundary, and can't really exist without these.
# Boundary conditions cannot be shared between profiles or between
# boundaries.
class BC(registeredclass.RegisteredClass):
    registry = []
    def __init__(self, profile, boundary, user_enable=1):

        self.profile = profile     # Profile object.
        self.boundary = boundary   # Name of the boundary.

        # Unpack the NamedProfile, if you got one.
        self.profile_obj = self.profile.unpack()

        self.mesh_obj = None

        self.sw_cb = [switchboard.requestCallback(
            "new namedprofile obj", self.replace_profile)]

        # BC's start out inactive, but are auto-activated at
        # add-to-mesh time.  If they are ever explicitly inactivated
        # by the user, the explicit_inactive flag gets set, and
        # auto_activate will not make them active.
        self.enabled = None
        self.explicit_enable = user_enable

    # Status-setting function -- enables the BC if it's possible to do
    # so and the user has never said not to.  This function should be
    # called whenever the fields/equations/flux activation info in the
    # mesh changes.  Subclasses must have the "computable" function,
    # which answers the "do you make sense" question, independently of
    # whether the BC will actually be enabled.
    def auto_enable(self):
        if not self.explicit_enable:
            self.enabled = None
            return
        self.enabled = self.computable()

    def is_enabled(self):
        return self.enabled


    # TODO: Users should be able to explicitly enable and disable BCs
    # via a menu item (and also the GUI, of course), and disabled BCs
    # should not be checked for collisions or promotion, so that users
    # can then switch back and forth between different candidate BCs
    # for their problem.
    
    # Functions for explicit enable or disable of BCs by users.
    def enable(self):
        self.explicit_enable = 1
        self.auto_enable()

    def disable(self):
        self.explicit_enable = None
        self.auto_enable()

    # Some BCs need to indicate if they're the result of promoting
    # other BCs.
    def promotion(self):
        return None

    def __eq__(self, other):
        return id(self)==id(other)
    
    # Add_to_mesh is separate from the construction process for the
    # convenience of the boundary-condition editing, which is
    # accomplished by creating a new boundary condition and adding it
    # to the mesh with the name of a recently-removed one. 
    def add_to_mesh(self, name, mesh, promotion=0):
        self.name = name 
        self.mesh = mesh # Mesh name.
        
        self.mesh_obj = oof2.engine.mesh.meshes[mesh] # Who object.
        self.femesh = self.mesh_obj.femesh()     # Actual mesh.

        try:
            self.boundary_obj = self.mesh_obj.getBoundary(self.boundary)
        except KeyError:
            debug.fmsg(self.boundary)
            raise

        self.profile_obj.addCondition(self)
        self.boundary_obj.addCondition(self)

        # Auto-enable must precede the mesh_obj add call below, which
        # builds the display string, which depends on side-effects of
        # auto_enable.
        self.auto_enable()

        self.mesh_obj.addBdyCondition(name, self, promotion=promotion)

        
    def check(self, mesh, exclude=None):
        # Check that this bc doesn't conflict with any other bc on
        # this mesh. This doesn't check the *values* of the boundary
        # conditions.  It only checks that two Dirichlet or floating
        # boundary conditions don't apply to the same degrees of
        # freedom on the same boundary.
        # If the 'exclude' variable is equal to some other bc, then
        # conflicts with that bc aren't counted.
        
        mesh_obj = oof2.engine.mesh.meshes[mesh]
        boundary_obj = mesh_obj.getBoundary(self.boundary)
        return boundary_obj.checkCondition(self, exclude=exclude)

    def remove_from_mesh(self, promotion=0):
        self.mesh_obj.rmBdyCondition(self.name, promotion=promotion)
        # rmBdyCondition calls disconnect, below.


    # Called prior to deletion, undo the additions we did on __init__.
    # Because of the circular references, as well as control over the
    # timing, these operations are unsuitable for a __del__ operation.
    def disconnect(self):
        self.profile_obj.removeCondition(self)
        self.boundary_obj.removeCondition(self)
        # Break circular references.
        self.boundary = None
        self.boundary_obj = None
        self.profile=None
        self.profile_obj = None
        # Clean up callbacks.
        map(switchboard.removeCallback, self.sw_cb)


    # Called from the NamedProfile object when profiles are edited.
    # The NamedProfile has a new .obj, and the conditions must all
    # update accordingly.
    def replace_profile(self, profile):
        if self.profile == profile:
            self.profile_obj = self.profile.unpack()
            self.profile_obj.addCondition(self)

    # Used by the GUI.
    def display_string(self):
        resstr = self.boundary_obj.name() + " / " + self.display() + \
                 " / %s" % self.profile_obj.descriptor()
        if self.enabled:
            return resstr
        else:
            return "(%s)" % resstr
        
##    def display_string(self):
##        bdyprofstring = " on boundary %s with profile %s." % \
##                        (`self.boundary_obj.name()`,
##                         self.profile_obj.descriptor())
##        return self.display() + bdyprofstring

    tip = "Different kinds of boundary conditions."
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/bc.xml')

# Individual boundary condition classes take the field and equation
# data in the form of a component object, which they must then change
# to a separate field and component for actual evaluation.  Boundary
# conditions are also specific to a particular mesh and boundary.
# They also know about a profile -- the object passed in may either be
# an existing profile in the current mesh, or a new unnamed profile.
# In either case, the passed in object is the appropriate subclass of
# ProfileFinder.

class DirichletBC(BC):
    def __init__(self, field, field_component, equation, eqn_component,
                 profile, boundary, 
                 offset=0.0, promotion = None, user_enable = 1):
        self.field = field
        self.field_component = field_component
        self.equation = equation
        self.eqn_component = eqn_component

        # Parent class __init__ needs to be last, because it does
        # collision detection on the boundary, which means the field
        # and equation have to already be set.
        BC.__init__(self, profile, boundary, user_enable=user_enable)

        # When a DirichletBC is auto-generated from a FloatBC, it can
        # acquire an offset, and gets "promoted" set.  
        self.offset = offset
        self.promoted = promotion

    # Fixed BCs need to indicate if they're the result of promoting
    # a floatBC.
    def promotion(self):
        return self.promoted
    
    # "Copy" makes a new BC with a new name on a new boundary, but
    # with the same other parameters, except explicit-enabling.
    def copy(self, boundary):
        return DirichletBC(self.field, self.field_component,
                           self.equation, self.eqn_component,
                           self.profile, boundary)

    # Dispatch routine to call the right boundary-addition gizmo.
    def addToBoundary(self, boundary):
        boundary.addFixedCondition(self)
    def checkBoundary(self, boundary, exclude):
        return boundary.checkFixedCondition(self, exclude)

    # Am I computable on the current mesh?
    def computable(self):
        if not self.mesh_obj:
            return None
        return self.femesh.is_defined_field(self.field) and \
               self.femesh.is_active_field(self.field) and \
               self.femesh.is_active_equation(self.equation)
        
        
    # BC is applied through its "call" method, which just sets
    # the value of the appropriate DOF to the passed-in value, plus
    # the offset, and marks the corresponding equation as
    # dependent.
    def __call__(self, node, value):
        if not self.enabled:
            return
        fldcomp = self.field.getIndex(self.field_component).integer()
        self.field.setvalue(self.femesh, node, fldcomp, value + self.offset )
        dof = self.field.dof(self.femesh, node, fldcomp)
        dof.fix()

        eqncomp = self.equation.getIndex(self.eqn_component).integer()
        nodalEqn = self.femesh.map_forward_equation(
            self.equation, node, eqncomp)
        nodalEqn.dependent()

    def intersect(self, other): # Other should be another DirichletBC or a FloatBC.
        return self.field == other.field and \
               self.field_component == other.field_component and \
               self.equation == other.equation and \
               self.eqn_component == other.eqn_component

    def display(self):
        return "DirichletBC / %s[%s] / %s[%s]" % (`self.field`,
                                              self.field_component,
                                              `self.equation`,
                                              self.eqn_component)
    
registeredclass.Registration("Dirichlet",
                             BC, DirichletBC,
                             ordering=1,
                             params = bc_field_param_group +
                             bc_equation_param_group + [ 
    parameter.RegisteredParameter(
    'profile', reg=profile.Profile, value=None,
    tip='Profile describing how this condition varies.'),
    bc_boundary_param
    ],
                             tip="Fix the values of a Field along a boundary.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/dirichlet.xml')
                             ) 


class ForceBC(BC):
    def __init__(self, equation, eqn_component, profile, boundary,
                 user_enable=1):
        self.equation = equation
        self.eqn_component = eqn_component

        BC.__init__(self, profile, boundary,user_enable=user_enable)

    def copy(self, boundary):
        return ForceBC(self.equation, self.eqn_component,
                       self.profile, boundary)
    #
    # def write(self, file, indent):
    #     file.write(indent+"ForceBC("+problem.inversedict[self.equation]+",")
    #     file.write("%d)" % self.equationComponent)
    #
    def addToBoundary(self, boundary):
        boundary.addForceCondition(self)
    def checkBoundary(self, boundary, exclude):
        return boundary.checkForceCondition(self, exclude)

    def computable(self):
        if not self.mesh_obj:
            return None
        return self.femesh.is_active_equation(self.equation)
    #
    # Applied via "call" method, just creates an applicator if it
    # doesn't already have one, and calls it with the node and
    # equation component, and the value to set.  
    #
    def __call__(self, node, value):
        if not self.enabled: return
        eqnindex = self.equation.getIndex(self.eqn_component).integer()
        boundarycond.applyForceBC(self.femesh, self.equation,
                                  node, eqnindex, value)
    def display(self):
        return "ForceBC / %s" % `self.equation`



registeredclass.Registration("Generalized Force",
                             BC,
                             ForceBC,
                             ordering=2,
                             params = bc_equation_param_group + [
    parameter.RegisteredParameter(
    'profile', reg=profile.Profile, value=None,
    tip='Profile describing how this condition varies.'),
    bc_point_bdy_param],
                             tip="Apply a force to all nodes on a boundary.  A force is a divergence of a flux (eg, stress, heat flux).  The condition is applied independently at each node, and is therefore sensitive to node density.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/forcebc.xml')
                             )


## Float boundary condition -- similar to a DirichletBC in basic
## structure, but with major differences in the apply function.  What
## 'apply' does is go through the mapping lists passed to the mesh by
## the matrix extraction routine, and set them up so that all the rows
## and columns corresponding to the DOFs and nodalequations of the
## floating boundary condition are remapped to a single row/colum,
## respectively, of the new mapping.

## They also have an unravelling routine which unmaps the DOFs
## after solution.

## The 'profile' function has the same structure but a different
## meaning here -- it's an offset, the true fully-solved value of the
## DOF is the profile value plus another value, such that the sum
## equals the sum arrived at by the solver.


class FloatBC(BC):
    def __init__(self, field, field_component, equation, eqn_component,
                 profile, boundary, user_enable=1):
        self.field = field
        self.field_component = field_component
        self.equation = equation
        self.eqn_component = eqn_component
        BC.__init__(self, profile, boundary, user_enable=user_enable)
        self.reset()

    def copy(self, boundary):
        return FloatBC(self.field, self.field_component,
                       self.equation, self.eqn_component,
                       self.profile, boundary)
        
    # This function can be called either by init or from the boundary,
    # when it wants to reset stateful boundary conditions.  This
    # provides a handy way of pretend-encapsulating the state variables.
    # FloatBCs are stateful because intersecting combinations of them
    # need to aggregate their degrees of freedom in a sensible way,
    # with consistent profiles and without double-counting.  This is
    # what the traversal process does.
    def reset(self):
        #
        # Below here are the "stateful" data members.  
        self.applicator = None     # C applicator -- "None" means first-call.
        self.dofmappingIndex = -1  # Set either locally or by parents
        self.eqnmappingIndex = -1  #    on the intersection tree.
        self.expandfirst = None    # Detects first call to "expand"
        self.offset = 0.0          # Local profile offset relative to parent.
        self.profileStart = 0.0    # Master start value, set by root.
        self.profileOffset = 0.0   # Cumulative total offset from intersections
        self.active = 1            # 
        self.dupnodes = {}         # Rationalized duplicate set.
        self.nodeshares = {}       # BC-key'd dict, parent : dict of dupnodes.
        self.intersectors = {}     # BC-key'd dict, child : offset
        self.intersected = None    # Intersection state.
        # States are None, "mapped", and "traversed."


        # A FloatBC is active by default, but can be deactivated
        # if it gets promoted to a dirichlet BC.  This happens when
        # the "fix" method is called.
        # The floatappset member helps to encode state information -- the
        # applicator knows its first call, which is why it's important
        # to de-allocate it at the end.
        #   "Activeness" is checked for locally, so that active and
        # inactive floatBC's present the same interface, but inactive
        # ones don't actually do anything.  The only other place
        # activeness is checked for is in the intersection routine.
        #
        # Duplicate nodes can arise from intersection -- the calls
        # simply do nothing at all in that case.  The first node of
        # the parent of an intersection tree is guaranteed not to
        # be duplicate to the parent, so all the order-depdendent
        # stuff still works.


    # Dispatch routine, called from the boundary, calls back to the
    # correct boundary-addition routine.
    def addToBoundary(self, boundary):
        boundary.addFloatCondition(self)
    def checkBoundary(self, boundary, exclude):
        return boundary.checkFloatCondition(self, exclude)
    #

    def computable(self):
        if not self.mesh_obj:
            return None
        return self.femesh.is_defined_field(self.field) and \
               self.femesh.is_active_field(self.field) and \
               self.femesh.is_active_equation(self.equation)
        
    
    def __call__(self, node, value):
        if not self.enabled: return
        # Statefulness is in the floatappset list -- if it's null, this is
        # the first call, you need to set up one for each field component.
        if self.active and not self.dupnodes.has_key(node):
            fldcomp = self.field.getIndex(self.field_component).integer()
            eqncomp = self.equation.getIndex(self.eqn_component).integer()
            value+=self.profileOffset
            if self.applicator:
                # This is a non-first call, just do editmap.
                self.applicator.editmap(self.femesh, value, node,
                                        self.field, fldcomp,
                                        self.equation, eqncomp,
                                        self.dofmappingIndex,
                                        self.eqnmappingIndex,
                                        self.profileStart)
            else:
                # This is the first call to this particular FloatBC.
                self.applicator = boundarycond.FloatBCApp()
                if not self.intersected:
                    # We are either the root of an intersection tree,
                    # or unintersected -- find out what mapping targets to use.
                    self.profileStart = value

                    self.dofmappingIndex = \
                      self.applicator.getdofindex(self.femesh, node,
                                                  self.field, fldcomp)

                    self.eqnmappingIndex = \
                        self.applicator.geteqnindex(
                        self.femesh, node, self.equation, eqncomp)

                    # Broadcast these to the intersecting set.
                    for intersector in self.intersectors.keys():
                        intersector.propagate_(self.dofmappingIndex,
                                               self.eqnmappingIndex,
                                               self.profileStart)
                else:
                    # If we're here, it means this is the first call
                    # to a non-root member of an intersection tree,
                    # so we need to call editmap on the first node,
                    self.applicator.editmap(
                        self.femesh, value, node,
                        self.field, fldcomp,
                        self.equation, eqncomp,
                        self.dofmappingIndex,
                        self.eqnmappingIndex,
                        self.profileStart)

    # This is the function that handles the propagation of parameters
    # through the intersections -- recursively propagates everything.
    def propagate_(self, dofmappingIndex, eqnmappingIndex, profileStart):
        if self.intersected != 'traversed':
            raise ooferror.ErrPyProgrammingeError(
                "Data propagated to incomplete Float.")
        self.dofmappingIndex = dofmappingIndex
        self.eqnmappingIndex = eqnmappingIndex
        self.profileStart = profileStart
        for intersector in self.intersectors.keys():
            intersector.propagate_(dofmappingIndex, eqnmappingIndex,
                                  profileStart)
            
    # "Enabled" indicates whether the fields and equations are
    # defined, or whether or not the BC has been turned off by the
    # user.  "Active" refers to whether or not this BC has survived
    # the promotion process.  They are orthogonal.
    def contrib_rhs(self):
        if not self.enabled: return
        if self.active:
            self.applicator.profile_rhs(self.femesh)
            self.applicator = None  # Null this, anticipating next round.
            self.expandfirst = None

    # After matrix solution, need to fill in the DOF values
    # according to the profile function.  Tuple length-matches not
    # checked, because if they're wrong, we failed in "invoke."
    # This routine breaks if the nodes don't come in the
    # same order as they did in invoke.  (More specifically, the same
    # node must occur first.)

    def expand(self, node, value):
        if not self.enabled: return
        # print "FloatBC::expand."
        if self.active and not self.dupnodes.has_key(node):
            value += self.profileOffset
            comp = self.field.getIndex(self.field_component).integer()
            if not self.intersected:
                if self.expandfirst==None:
                    self.expandfirst = 1
                else:
                    prev = self.field.value(self.femesh, node, comp)
                    # print "Assigning component ", comp, " of field ", self.field_component, " the value ", value-self.profileStart_prev, " at node ", node
                    self.field.setvalue(self.femesh, node, comp,
                                        value-self.profileStart+prev)
            else: # Intersection case -- profileStart assumed received.
                prev = self.field.value(self.femesh, node, comp)
                self.field.setvalue(self.femesh, node, comp,
                                    value-self.profileStart+prev)
                    

    # Build the equivalent DirichletBC for your field and equation,
    # and deactivate yourself.
    def fix(self, offset):
        self.active = 0
        return DirichletBC(
            self.field, self.field_component,
            self.equation, self.eqn_component,
            self.profile, self.boundary, offset=offset, promotion=1,
            enabled=self.enabled)
        
    #
    def intersect(self, other): # Other should be another FloatBC.
        return self.field == other.field and \
               self.field_component == other.field_component and \
               self.equation == other.equation and \
               self.eqn_component == other.eqn_component

    def display(self):
        return "FloatBC / %s[%s] / %s[%s]" % (`self.field`,
                                              self.field_component,
                                              `self.equation`,
                                              self.eqn_component)

registeredclass.Registration("Floating",
                             BC,
                             FloatBC,
                             ordering=3,
                             params= bc_field_param_group +
                             bc_equation_param_group + [
    parameter.RegisteredParameter(
    'profile', reg=profile.Profile, value=None,
    tip='Profile describing how this condition varies.'),
    bc_boundary_param
    ],
                             tip="Fix the values of Fields on the boundary, up to an unspecified offset.",
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/floating.xml')
                             )



########################################################################
# # # # # # # # # # NEUMANN BOUNDARY CONDITION # # # # # # # # # # # # #
########################################################################

from oof2.engine.profile import Location
from oof2.SWIG.common import coord

# Wrapper function for flux boundary profiles.  Takes a number of
# floating-point values provided from within C++, translates them into
# a Location object, and calls the profile object with the location.
# OPT: If this is too slow, the Location construction process can be
# moved into the interior of the flux class's BCCallback routines,
# defined in boundarycond.swg.
def flux_locator(profile_obj, posx, posy, time, normx, normy,
                 distance, fraction):
    pos = coord.Coord(posx, posy)
    norm = coord.Coord(normx, normy)
    loc = Location(pos, normal=norm, s=distance, alpha=fraction, time=time)
    return profile_obj(loc)

#
# Actual flux boundary condition.  Simpler than the others, because
# most of the action happens in C++.
class NeumannBC(BC):
    def __init__(self, flux, profile, boundary, normal = None,
                 user_enable=1):
        self.flux = flux
        self.normal = normal  
        BC.__init__(self, profile, boundary, user_enable=user_enable)

    def copy(self, boundary):
        return NeumannBC(self.flux, self.profile, boundary,
                         normal=self.normal)
    
    def write(self, file, indent):
        file.write(indent+"NeumannBC("+problem.inversedict[self.flux]+")")
    def addToBoundary(self, boundary):
        boundary.addFluxCondition(self)
    def checkBoundary(self, boundary, exclude):
        return boundary.checkFluxCondition(self, exclude)

    def computable(self):
        if not self.mesh_obj:
            return None
        return self.femesh.is_active_flux(self.flux)
    
    #
    # Invoke means to do the integral over the appropriate
    # boundary, and make the contributions to the RHS.
    def __call__(self, edgeset, profile, time):
        if not self.enabled: return
        if self.femesh.is_active_flux(self.flux):
            applicator = boundarycond.NeumannBCApp(self.femesh, self.flux, edgeset)
            # For this call, "profile" must be a Python callable.
            applicator.integrate(flux_locator, profile, self.normal, time)
        else:
            raise ErrSetupError('Attempt to invoke NeumannBC on inactive flux.')
    def display(self):
        return "NeumannBC/%s" % `self.flux`
##    def display(self):
##        return "NeumannBC(flux=%s)" % `self.flux`

# NeumannBC doesn't make sense to apply point-wise, use the ForceBC for
# that.  ForceBC's are specified in terms of equations and components,
# rather than fluxes, because it makes more sense. 


registeredclass.Registration(
    "Neumann",
    BC,
    NeumannBC,
    ordering=4,
    params = [meshparameters.FluxParameter("flux",
                                           tip="Flux whose value is specified across this boundary."),

              profile.FluxProfileSetParameter("profile",
                                              tip="Profiles describing the prescribed flux."),
              bc_edge_bdy_param,
              parameter.BooleanParameter("normal",
                                         tip="If true, the boundary condition is given in a local right-handed coordinate system where x is the outward normal to the boundary and y is the tangent direction.  If false, x and y are lab coordinates."
                                         )],
    tip="Set the normal component of a Flux (eg, stress, heat flux) along the boundary.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/neumann.xml'))

