// -*- C++ -*-
// $RCSfile: thermalexpansion.C,v $
// $Revision: 1.39 $
// $Author: langer $
// $Date: 2005/06/15 21:14:39 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>


#include "common/coord.h"
#include "engine/IO/propertyoutput.h"
#include "engine/element.h"
#include "engine/equation.h"
#include "engine/femesh.h"
#include "engine/flux.h"
#include "engine/material.h"
#include "engine/matvec.h"
#include "engine/property/elasticity/cijkl.h"
#include "engine/property/elasticity/elasticity.h"
#include "engine/IO/propertyoutput.h"
#include "engine/cstrain.h"
#include "engine/property/orientation/orientation.h"
#include "thermalexpansion.h"


ThermalExpansion::ThermalExpansion(PyObject *reg, 
				   const std::string &nm, 
				   double t0)
  : Property(nm,reg),
    tzero_(t0)
{
  temperature=dynamic_cast<ScalarField*>(Field::getField("Temperature"));
  stress_flux=dynamic_cast<SymmetricTensorFlux*>(Flux::getFlux("Stress"));
}


void ThermalExpansion::precompute() {

}

int ThermalExpansion::K_integration_order(int degree) const {
  return 2*degree - 1;
}


void ThermalExpansion::fluxmatrix(const FEMesh *mesh,
				  const Element *element,
				  const ElementFuncNodeIterator &nu, 
				  const Flux *flux, FluxData *fluxdata,
				  const MasterPosition &x) const
{
  // The stress-free strain is: 
  //                    [ alpha11*T   0      0  ]
  //     u_{kl}^0(T) =  [    0    alpha22*T  0  ]
  //                    [    0    0   alpha33*T ]
  //
  // Our goal is to subtract C_{ijkl}*u_{kl}^0(T) from the stress
  // tensor.
  if(*flux != *stress_flux) {
    throw ErrProgrammingError("Unexpected flux." __FILE__, __LINE__);
  }
  
//   Trace("ThermalExpansion::fluxmatrix");

  // Assume linear elasticity, for now. 
  const Cijkl modulus = elasticity->cijkl(mesh, element, x);

  double sfval = nu.shapefunction(x);
  // Loop over all tensor components.
  for(SymTensorIterator ij; !ij.end(); ++ij) {
    double &mtx_el = fluxdata->matrix_element(mesh, ij, temperature, nu); //returns flux matrix
    // Strain tensor components.
    for(SymTensorIterator kl; !kl.end(); ++kl) {
      double ca = modulus(ij, kl)*expansiontensor(kl.row(), kl.col())*sfval;
      //std::cerr << "alphaDT=" << ca << std::endl;
      if(kl.diagonal()) {
	mtx_el -= ca;
      }
      else {
	mtx_el -= 2.0*ca;
      }
    }
  }
}

void ThermalExpansion::fluxrhs(const FEMesh *mesh, const Element *element,
				   const Flux *flux, FluxData *fluxdata,
				   const MasterPosition &x) const {
  
  if(*flux!=*stress_flux) {
    throw ErrProgrammingError("Unexpected flux." __FILE__, __LINE__);
  }
  const Cijkl modulus = elasticity->cijkl(mesh, element, x);

  for(SymTensorIterator ij; !ij.end(); ++ij) {
    double &offset_el = fluxdata->offset_element(mesh, ij); // reference!
    for(SymTensorIterator kl; !kl.end(); ++kl) {
      if(kl.diagonal()) {
	offset_el += modulus(ij,kl)*expansiontensor(kl.row(),kl.col())*tzero_;
      }
      else {
	offset_el += 2.0*modulus(ij,kl)*expansiontensor(kl.row(),kl.col())*tzero_;
      }
    }
  }
}


void ThermalExpansion::output(const FEMesh *mesh,
				  const Element *element,
				  const PropertyOutput *output,
				  const MasterPosition &pos,
				  OutputVal *data)
  const 
{
  const std::string &outputname = output->name();
  if(outputname == "Strain") {
    // The parameter is a Python StrainType instance.  Extract its name.
    acquirePyLock();
    std::string stype;
    try {
      PyObject *str = PyObject_CallMethod(output->parameters()[0], "name", 0);
      stype = PyString_AsString(str);
      Py_XDECREF(str);
    }
    catch(...) {
      releasePyLock();
      throw;
    }
    releasePyLock();
    SymmMatrix3 *sdata = dynamic_cast<SymmMatrix3*>(data);
    // Compute alpha*T with T interpolated to position pos
    // TODO LATER: This is baroque and may be slow.  Make a
    // specialized function for getting the value of a scalar field.
    const OutputValue tfield = element->outputField(mesh, *temperature, pos);
    const ScalarOutputVal *tval = 
      dynamic_cast<const ScalarOutputVal*>(tfield.valuePtr());
    double t = tval->value();
    

    if(stype == "Thermal")
      *sdata += expansiontensor*(t-tzero_);
    else if(stype == "Elastic")
      *sdata -= expansiontensor*(t-tzero_);
  }//strain output ends here
  if(outputname == "Energy") {
    // The parameter is a Python Enum instance.  Extract its name.
    std::string etype;
    acquirePyLock();
    try {
      PyObject *str = PyObject_GetAttrString(output->parameters()[0], "name");
      etype = PyString_AsString(str);
      Py_XDECREF(str);
    }
    catch(...) {
      releasePyLock();
      throw;
    }
    releasePyLock();
    if(etype == "Total" || etype == "Elastic") {
      ScalarOutputVal *edata = 
	dynamic_cast<ScalarOutputVal*>(data);
      SymmMatrix thermalstrain(3);
      const Cijkl modulus = elasticity->cijkl(mesh, element, pos); //returns lab reference system stiffness
      const OutputValue tfield = element->outputField(mesh, *temperature, pos);
      const ScalarOutputVal *tval = 
	dynamic_cast<const ScalarOutputVal*>(tfield.valuePtr());
      double t = tval->value();
      thermalstrain = expansiontensor*(t-tzero_);
      SymmMatrix thermalstress(modulus*thermalstrain);
      SymmMatrix strain(3);
      findGeometricStrain(mesh, element, pos, strain);
      double e = 0;
      for(int i=0; i<3; i++) {
	e += thermalstress(i,i)*strain(i,i);
	int j = (i+1)%3;
	e += 2*thermalstress(i,j)*strain(i,j);
      }
      *edata += -0.5*e;
    }
  }//energy output ends here
}


int ThermalExpansion::is_symmetric(const FEMesh* mesh) const {
  std::vector<Equation*> &eqns = Equation::alleqns();
  for (int i =0; i<eqns.size(); i++)
    {
      if (eqns[i]->name() == "Force_Balance")
	{
	  if (eqns[i]->is_active(mesh))
	    if (temperature->is_defined(mesh) && temperature->is_active(mesh))
	      return 0;
	}
    }
  return 1;

}

IsotropicThermalExpansion::IsotropicThermalExpansion(PyObject *registry,
						     const std::string &name, 
						     double alpha,
						     double t0)
  : ThermalExpansion(registry, name, t0),
    alpha_(alpha)
{

}


void IsotropicThermalExpansion::cross_reference(Material *mat) {
  // find out which property is the elasticity
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
  
}

void IsotropicThermalExpansion::precompute() {
  ThermalExpansion::precompute();
  expansiontensor(0,0) = expansiontensor(1,1)
    = expansiontensor(2,2) = alpha_;
}

AnisotropicThermalExpansion::AnisotropicThermalExpansion(PyObject *registry, 
							 PyObject *self,
							 const std::string &nm, 
							 SymmMatrix3 *alpha, double t0)
  :  alpha_(*alpha),
     ThermalExpansion(registry, nm, t0),
     PythonNative<Property>(self) 
{

}
void AnisotropicThermalExpansion::cross_reference(Material *mat) {
  // find out which property is the elasticity
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
  orientation = dynamic_cast<OrientationProp*>
    (mat->fetchProperty("Orientation"));
}

void AnisotropicThermalExpansion::precompute()
{
  expansiontensor = alpha_.transform(orientation->eulerangle());
}
